// -*- C++ -*-
#include "AGILe/CCHerwig/CCHerwig.hh"
#include "ThePEG/Repository/EventGenerator.h"
#include "ThePEG/Repository/Repository.h"
#include "ThePEG/Utilities/DynamicLoader.h"
#include "ThePEG/EventRecord/Event.h"
#include "ThePEG/Vectors/HepMCConverter.h"


namespace ThePEG {
  template<> 
  struct HepMCTraits<HepMC::GenEvent> 
    : public HepMCTraitsBase<HepMC::GenEvent,
                             HepMC::GenParticle,
                             HepMC::GenVertex,
                             HepMC::Polarization> 
  {};
}
using namespace ThePEG;


namespace AGILe {

  // Authors: Lars Sonnenschein, 2007-02-13
  //          Andy Buckley, 2007-10-18

  // Standard constructor
  CCHerwig::CCHerwig() {
    breakThePEG();
    const string libpath = HWLIBDIR;
    const string repopath = HWDATADIR;
    DynamicLoader::prependPath(libpath);
    InterfaceBase::NoReadOnly = true;
    Repository::read(repopath + "/HerwigDefaults.in", cout);
    Repository::update();
    InterfaceBase::NoReadOnly = false;
    Repository::exec("mkdir /Agile" , cout);
    Repository::exec("cd /Agile" , cout);

    // Make subprocess handler (needs to be initial-state specific)
    // For e+ e-:
    Repository::exec("create ThePEG::SubProcessHandler AgileSubProc" , cout);
    Repository::exec("insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEee2gZ2qq" , cout);
    // For hadron-hadron (by default only gamma/Z switched on)
    // insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEqq2gZ2ff
    // #insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEqq2W2ff
    // #insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEGammaGamma
    // #insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEGammaJet
    // #insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEHiggs
    // #insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEHiggsJet
    // #insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEQCD2to2
    // #insert AgileSubProc:MatrixElements[0] /Defaults/MatrixElements/MEHeavyQuark

    // Make "extractor"
    Repository::exec("cp /Defaults/Handlers/StandardExtractor AgileExtractor", cout);
    Repository::exec("set AgileSubProc:PartonExtractor AgileExtractor" , cout);

    // Make general event handler
    Repository::exec("create ThePEG::StandardEventHandler AgileHandler", cout);
    Repository::exec("create ThePEG::LuminosityFunction AgileLuminosity", cout);
    Repository::exec("insert AgileHandler:SubProcessHandlers[0] AgileSubProc", cout);
    Repository::exec("set AgileHandler:Sampler /Defaults/ACDCSampler", cout);
    Repository::exec("set AgileHandler:CascadeHandler /Defaults/Handlers/Shower/stdShowerHandler", cout);
    Repository::exec("set AgileHandler:HadronizationHandler /Defaults/Handlers/Hadronization/Cluster/stdCluHadHandler", cout);
    Repository::exec("set AgileHandler:DecayHandler /Defaults/Handlers/DecayHandler", cout);
    // set AgileHandler:BinStrategy 2
    // Uncomment to get QED radiation off W/Z decay products in hard process:
    // #insert AgileHandler:PostSubProcessHandlers[0] /Defaults/Handlers/QED/QEDRadiationHandler

    Repository::exec("create ThePEG::EventGenerator AgileGenerator" , cout);
    Repository::exec("set AgileGenerator:RandomNumberGenerator /Defaults/Random" , cout);
    Repository::exec("set AgileGenerator:StandardModelParameters /Defaults/Model" , cout);

    Repository::exec("create ThePEG::Cuts AgileCuts", cout);
    // Cut on the CMS energy for photon radiation
    Repository::exec("set AgileCuts:MHatMin 22.36", cout);
    // # create the cuts object for hadron collisions
    // create ThePEG::Cuts QCDCuts
    // # cuts on all process
    // set QCDCuts:ScaleMin 1.0*GeV
    // set QCDCuts:X1Min 0.00001
    // set QCDCuts:X2Min 0.00001
    // set QCDCuts:MHatMin 20.
    // # cut on jet pt 
    // create ThePEG::SimpleKTCut JetKtCut SimpleKTCut.so
    // set JetKtCut:Matcher /Defaults/Matchers/MatchStandardQCDParton
    // set JetKtCut:MinKT 20.0*GeV
    // # cut on photon
    // create ThePEG::SimpleKTCut PhotonKtCut SimpleKTCut.so
    // set PhotonKtCut:Matcher /Defaults/Matchers/MatchPhoton
    // set PhotonKtCut:MinKT 20.0*GeV
    // # cut on leptons
    // create ThePEG::SimpleKTCut LeptonKtCut SimpleKTCut.so
    // set LeptonKtCut:Matcher /Defaults/Matchers/MatchLepton
    // set LeptonKtCut:MinKT 0.0*GeV
    // #cut on top quarks
    // create ThePEG::SimpleKTCut TopKtCut SimpleKTCut.so
    // set TopKtCut:Matcher /Defaults/Matchers/MatchTop
    // set TopKtCut:MinKT 0.0*GeV
    // # create a cut on the invariant mass of lepton pairs
    // create ThePEG::V2LeptonsCut MassCut V2LeptonsCut.so
    // set MassCut:MaxM 14000*GeV
    // set MassCut:Families All
    // set MassCut:CComb All
    // set MassCut:MinM 20.*GeV
    // set MassCut:MaxM 14000.*GeV
    // # insert into hadron cuts
    // insert QCDCuts:OneCuts[0] JetKtCut
    // insert QCDCuts:OneCuts[1] PhotonKtCut
    // insert QCDCuts:OneCuts[2] LeptonKtCut
    // insert QCDCuts:OneCuts[3] TopKtCut
    // insert QCDCuts:MultiCuts[0] MassCut
    
    Repository::exec("set AgileGenerator:RandomNumberGenerator:Seed 31122001", cout);
    Repository::exec("set AgileGenerator:DebugLevel 1", cout);
    Repository::exec("set AgileGenerator:PrintEvent 0", cout);
    Repository::exec("set AgileGenerator:MaxErrors 10000", cout);
    Repository::exec("set AgileGenerator:NumberOfEvents 1000000000", cout);
  }


  // Set up initial state from supplied params
  Status CCHerwig::setGenSpecificInitialState(PdgCode p1, double e1, 
                                              PdgCode p2, double e2) {
    cout << "Setting initial state..." << endl;

    /// @todo Actually put some values in!
    Repository::exec("set AgileLuminosity:BeamEmaxA 40.0", cout);
    Repository::exec("set AgileLuminosity:BeamEmaxB 52.0", cout);
    Repository::exec("set AgileHandler:BeamA /Defaults/Particles/e-", cout);
    Repository::exec("set AgileHandler:BeamB /Defaults/Particles/e+", cout);

    _initialstateset = true;
    return SUCCESS;
  }
  

  // Run generator initialization
  Status CCHerwig::initialize() {
    Generator::initialize();
    cout << "Initialising..." << endl;
    
    Repository::exec("set AgileHandler:LuminosityFunction AgileLuminosity", cout);
    Repository::exec("set AgileGenerator:EventHandler AgileHandler" , cout);
    Repository::exec("set AgileGenerator:EventHandler:Cuts AgileCuts", cout);
    
    EGPtr tmp = BaseRepository::GetObject<EGPtr>("/Herwig/Generators/AgileGenerator");
    eg = Repository::makeRun(tmp, "Agile");
    eg->initialize();
    _initialized = true;
    return SUCCESS;
  }


  Status CCHerwig::setSeed(const int seed) {
    Generator::setSeed(seed);
    setParam("AgileGenerator:RandomNumberGeneratorSeed", seed);
    return SUCCESS;
  }


  // Set parameters.
  Status CCHerwig::setParam(const string& name, const string& value) {
    Generator::setParam(name, value);
    Repository::exec("set " + name + " " + value, cout);
    return SUCCESS;
  }

  
  // Run the generator for one event.
  Status CCHerwig::makeEvent(HepMC::GenEvent& evt) {
    Generator::makeEvent(evt);
    try {
      if (eg) {
        // Generate an event and convert to a HepMC::GenEvent
        EventPtr event = eg->shoot();
        HepMC::GenEvent* currentEvent = HepMCConverter<HepMC::GenEvent>::convert(*event);
        /// @todo Problem with concurrent GenEvents?
        evt = *currentEvent; // <- actual copy assignment, not reference re-pointing.
        delete currentEvent;
      } else {
        cerr << "eg = nil" << endl;
      }      
    }
    catch ( std::exception & e ) {
      cerr << e.what() << endl;
      return FAILURE;
    }
    return SUCCESS;
  }
  

  /// Tidy up, print out run stats, etc.
  Status CCHerwig::finalize() {
    cout << "Finalising..." << endl;
    assert(eg);
    eg->finalize();
    return SUCCESS;
  }


}


// Class factory
extern "C" {
  AGILe::Generator* create() { return new AGILe::CCHerwig(); }
  void destroy(AGILe::Generator* gen) { delete gen; }
}
