#include "AGILe/AGILe.hh"
#include "AGILe/Generator.hh"
#include <dlfcn.h>


namespace AGILe {

  // Declare getLibPath function (uses binreloc)
  const string getLibPath();


  namespace Loader {
    
    // Static library handles
    vector<void*> handles;
    void* agilehandle = 0;
    string version;
    

    // Initialise the dynamic loader
    void initialize() {
      // Nothing to do with raw dlopen
    }


    void* loadGenLib(string name, int dlmode=(RTLD_LAZY|RTLD_GLOBAL)) {
      vector<string> dirs;
      char* env = 0;
      // Put "lib" on the start if needed
      if (name.find("lib") != 0) name = "lib" + name;
      // Put lib suffix on the end if needed
      if (name.find(SYSDSO) == string::npos) name += SYSDSO;
      
      // Then use the AGILe gen path variable
      env = getenv("AGILE_GEN_PATH");
      if (env) dirs += split(env);

      // If there's a CMTCONFIG variable, try to automatically use the generator
      // libraries from the Genser area on CERN AFS (CMTCONFIG specifies the 
      // LCG platform tag, which is hard to obtain programmatically). Also try a 
      // fallback of the LCG tag, computed at build time.
      string genserpath = "/afs/cern.ch/sw/lcg/external/MCGenerators/.liblinks/";
      env = getenv("CMTCONFIG");
      if (env) dirs.push_back(genserpath + env + "/lib");
      env = getenv("LCG_PLATFORM");
      if (env) dirs.push_back(genserpath + env + "/lib");
      dirs.push_back(genserpath + LCGTAG + "/lib");

      // Always (try to) use the AGILe library install path
      dirs.push_back(getLibPath());

      // Use the current dir, too.
      dirs.push_back(".");      
      
      // And then the user's (non-system) library path
      env = getenv("LD_LIBRARY_PATH");
      if (env) dirs += split(env);

      // Try to load lib here
      for (vector<string>::const_iterator d = dirs.begin(); d != dirs.end(); ++d) {
        string libpath = *d + "/" + name;
        while (libpath.find("//") != string::npos) {
          libpath.replace(libpath.find("//"), 2, "/");
        }
        void* lib = dlopen(libpath.c_str(), dlmode);
        Log::getLog("AGILe.Loader") << Log::TRACE << "Trying to load: " << libpath << " -> " << lib << endl;
        if (lib) return lib;
      }
      return 0;
    }


    vector<string> getAvailableGens() {
      // Possible generators.
      map<string, vector<string> > gens;
      vector<string> reqlibs;

      {
        // Fortran Herwig + reqlibs
        set<string> versions;
        versions.insert("6507");
        versions.insert("6510");
        for (set<string>::const_iterator v = versions.begin(); v != versions.end(); ++v) {
          const string baselib = "libherwig" + *v;
          vector<string> baselibs;
          baselibs.push_back(baselib);
          baselibs.push_back(baselib + "_dummy");
          baselibs.push_back(baselib + "_pdfdummy");
          gens["FHerwig:" + *v] = baselibs;
          reqlibs = baselibs;
          reqlibs.push_back("libjimmy");
          gens["FHerwigJimmy:" + *v] = reqlibs;
          reqlibs = baselibs;
          reqlibs.push_back("libalpher");
          gens["AlpGenFHerwig:" + *v] = reqlibs;
          reqlibs = baselibs;
          reqlibs.push_back("libcharybdisher");
          gens["CharybdisFHerwig:" + *v] = reqlibs;
        }
      }

      {
        // Fortran Pythia + reqlibs
        set<string> versions;
        versions.insert("6206");
        versions.insert("6326");
        versions.insert("6403");
        versions.insert("6404");
        versions.insert("6406");
        versions.insert("6408");
        versions.insert("6411");
        versions.insert("6412");
        versions.insert("6413");
        for (set<string>::const_iterator v = versions.begin(); v != versions.end(); ++v) {
          const string baselib = "libpythia" + *v;
          vector<string> baselibs;
          baselibs.push_back(baselib);
          baselibs.push_back(baselib + "_dummy");
          baselibs.push_back(baselib + "_pdfdummy");
          gens["FPythia:" + *v] = baselibs;
          reqlibs = baselibs;
          reqlibs.push_back("libalppyt");
          gens["AlpGenFPythia:" + *v] = reqlibs;
          reqlibs = baselibs;
          reqlibs.push_back("libcharybdispyt");
          gens["CharybdisFPythia:" + *v] = reqlibs;
        }
      }

      // Sherpa
      reqlibs.clear();
      reqlibs.push_back("libSherpaMain");
      gens["Sherpa"] = reqlibs;

      // Pythia 8
      reqlibs.clear();
      reqlibs.push_back("libpythia8");
      reqlibs.push_back("libhepmcinterface");
      gens["CCPythia"] = reqlibs;

      // Herwig++
      reqlibs.clear();
      reqlibs.push_back("libHerwig");
      gens["CCHerwig"] = reqlibs;


      // Find which generators are present by trying to dlopen() them.
      vector<string> availableGens;
      for (map<string, vector<string> >::const_iterator g = gens.begin(); g != gens.end(); ++g) {
        // All listed libraries must be available for each generator to be considered useable.
        bool ok = true;
        for (vector<string>::const_iterator l = g->second.begin(); l != g->second.end(); ++l) {
          void* lib = loadGenLib(*l, RTLD_LAZY);
          if (lib) {
            dlclose(lib);
            continue;
          } else {
            ok = false;
            break;
          }
        }
        if (! ok) continue;
        availableGens.push_back(g->first);
      }
      return availableGens;
    }


    void loadGenLibs(const string& genname) {
      const string fioName = "libHepMCfio";
      string agileName = "";
      string pdflibName = "";

      // Get version string from colon separated version string if given.
      string name = genname;
      version = "";
      const size_t colonpos = genname.rfind(":");
      if (colonpos != string::npos) {
        const size_t versionlength = genname.length() - colonpos - 1;
        name = genname.substr(0, colonpos);
        version = genname.substr(colonpos+1, versionlength);
      }

      // Herwig 6
      if (genname.find("FHerwig") != string::npos) {
        handles.push_back(loadGenLib("libherwig" + version));
        if (genname.find("AlpGen") == 0) {
          const string alpgenFHerwigName = "libalpher";	
          handles.push_back(loadGenLib(alpgenFHerwigName));
          agileName = "libAGILeAlpGenFHerwig";
        } else if (genname.find("Charybdis") == 0) {
          handles.push_back(loadGenLib("libcharybdisher"));
          agileName = "libAGILeCharybdisFHerwig";
        } else if (genname.find("Jimmy") != string::npos) {
          handles.push_back(loadGenLib("libjimmy"));
          agileName = "libAGILeFHerwigJimmy";
        } else {
          agileName = "libAGILeFHerwig";
        }
        handles.push_back(loadGenLib("libherwig" + version + "_dummy"));
        pdflibName = "libherwig" + version + "_pdfdummy";
        handles.push_back(loadGenLib(fioName));
      }

      // Pythia 6
      if (genname.find("FPythia") != string::npos) {
        const string charybName = "libcharybdispyt";
        const string libname = "libpythia" + version;
        handles.push_back(loadGenLib(libname));
        if (genname.find("AlpGen") == 0) {
          const string alpgenFPythiaName = "libalppyt";
          handles.push_back(loadGenLib(alpgenFPythiaName));
          agileName = "libAGILeAlpGenFPythia";
        } else if (genname.find("Charybdis") == 0) {
          handles.push_back(loadGenLib(charybName));
          agileName = "libAGILeCharybdisFPythia";
        } else {
          agileName = "libAGILeFPythia";
        }
        const string dummylibname = "libpythia" + version + "_dummy";
        handles.push_back(loadGenLib(dummylibname));
        pdflibName = "libpythia" + version + "_pdfdummy";
        handles.push_back(loadGenLib(fioName));
      }

      // Pythia 8
      if (genname.find("CCPythia") == 0) {
        if (version == "") version = "8";
        const string libname = string("libpythia") + version;
        handles.push_back(loadGenLib(libname));
        const string hepmcintname = "libhepmcinterface";
        handles.push_back(loadGenLib(hepmcintname));
        agileName = "libAGILeCCPythia";
      }

      // Herwig++
      if (genname.find("CCHerwig") == 0) {
        const string libName = "libHerwig";
        handles.push_back(loadGenLib(libName));
        agileName = "libAGILeCCHerwig";
      }

      // Sherpa
      if (genname.find("Sherpa") == 0) {
        const string libName = "libSherpaMain";
        handles.push_back(loadGenLib(libName));
        agileName = "libAGILeSherpa";
      }

      // Load LHAPDF *after* the generator library.
      void* lhalib = loadGenLib("libLHAPDF");
      // Fall back to dummy PDFs if LHAPDF isn't available
      if (!lhalib && !pdflibName.empty()) {
        handles.push_back(loadGenLib(pdflibName));
      } else {
        handles.push_back(lhalib);
      }

      // Load the appropriate AGILe library.
      agilehandle = loadGenLib(agileName);
      handles.push_back(agilehandle);

      // Check and reset dlopen errors
      for (vector<void*>::const_iterator h = handles.begin(); h != handles.end(); ++h) {
        if (*h == 0) {
          throw runtime_error(string("Failed to load libraries: ") + dlerror());
        }
      }
      dlerror();
    }


    Generator* createGen() {
      // Load the factory and create an instance of the class
      Creator* create_gen = (Creator*) dlsym(agilehandle, "create");
      if (char* err = dlerror()) {
        throw runtime_error((string("Can't load creation factory: ") + err).c_str());
      }
      
      Generator* created = create_gen();
      created->setVersion(version);
      
      return created;
    }


    void destroyGen(Generator* gen) {
      //Destroyer* destroy_gen = (Destroyer*) dlsym(agilemod, "destroy");
      //if (dlerror()) {
      //  throw runtime_error("Can't load destruction factory");
      //}
      //destroy_gen(gen);
      delete gen;
    }


    // Close the libraries
    void finalize() {
      //for (vector<void*>::const_reverse_iterator h = handles.rbegin(); h != handles.rend(); ++h) {
      for (vector<void*>::const_iterator h = handles.begin(); h != handles.end(); ++h) {
        if (*h == 0) dlclose(*h);
      }
    }


  }
}
